// g++ --std=c++1z -Wall -Wextra -O3 -o ool_benchmark ool_benchmark.cpp
#include "out_of_line.h"
#include <chrono>
#include <iostream>
#include <thread>
#include <vector>

// Deomonstrates keeping the cold data in the object with the fast data
struct Together {
  std::uint32_t value;
  std::string metadata;
};

// Demonstrates "giving up" and just not storing the associated cold-data
struct OnlyFastData {
  std::uint32_t value;
};

// Demonstrates using OutOfLine to store the cold data
struct WithOOL : public OutOfLine<WithOOL, std::string> {
  std::uint32_t value;
};

// The crux of the optimization is that there's no space-overhead
static_assert(sizeof(WithOOL) == sizeof(OnlyFastData));

// We synthesize the data up-front so the generation can't interfere with the measured behavior in any way
template <class Data>
auto make_data() {
  srand(20180101);
  std::vector<Data> data(10000000);
  for (Data& d : data) { d.value = rand(); }
  return data;
}

// We time to measure throughput of touching all the fast data once in sequence
template <class Data>
void trial(const char* const name, const Data& data, const bool print) {
  std::uint32_t running = 0U;
  const auto before = std::chrono::system_clock::now();
  for (const auto& d : data) { running += d.value; }
  const auto after = std::chrono::system_clock::now();

  if (print) {
    std::cout << name << " took " << (after - before).count() << "ns and " << running
              << " is a value I don't want optimized away" << std::endl;
  }
}

int main() {
  // We generate all the datasets up front, to keep the effects of the allocator distant
  const auto together  = make_data<Together>();
  const auto only_fast = make_data<OnlyFastData>();
  const auto with_ool  = make_data<WithOOL>();

  // We run each trial twice, recording the results only the second time. The idea is to give a fair comparison, where
  // all 3 options have their caches primed and nobody thus has data fresher from the allocation.
  trial("With cold data in-line (original)               ", together, false);
  trial("With cold data in-line (original)               ", together, true);

  trial("With cold data thrown away (best-case scenario) ", only_fast, false);
  trial("With cold data thrown away (best-case scenario) ", only_fast, true);

  trial("With OutOfLIne                                  ", with_ool, false);
  trial("With OutOfLIne                                  ", with_ool, true);
return 0;
}
